#!/usr/bin/env python
# coding: utf-8

# In[ ]:


# -----------------------------------------------------------------------------
# Method for Bayesian Hypotheses testing superseeds the t-test (BEST)
# according to Kruschke 2012
# -----------------------------------------------------------------------------

import matplotlib.pyplot as plt
import arviz as az
import seaborn as sns
import scipy.stats as stats
import numpy as np
import pymc


class Hypothese_testing:
    """Class for hypothese testing of two datasets.
    The data is assinged to the class and bayesian hypotheses testing methods
    can be applied.
    Input:
        data1 : First data set (numpy array)
        data2 : Second data set to compare against
        data1_name : name of the first dataset (used for the plot titles)
        data2_name : name of the second dataset
    """

    def __init__(self, data1, data2, data1_name: str,
                 data2_name: str, compare_name: str):
        """Data of the hypotheses that are tested.
        Input:
            data1 : first dataset
            data2 : second dataset to compare against
            compare_name : name of the test that was made
                           (included in file names)
        """

        self.D1 = data1
        self.D2 = data2
        self.D1_name = data1_name
        self.D2_name = data2_name
        self.compare_name = compare_name

        self.D1_n = len(data1)  # sample size
        self.D2_n = len(data2)

        # Test if data is in the right format
        if not isinstance(self.D1,(np.ndarray, np.generic)) and isinstance(self.D2,(np.ndarray, np.generic)):
            raise TypeError('Hypothesis data must be of type numpy ndarray.')

    def histogram_compare(self):
        """Plots seaborn histogram of the two original datasets.
        The data of the class is used."""

        plt.rcParams['figure.figsize'] = (6, 3.5)  # Set global parameters
        plt.rcParams['xtick.labelsize'] = 17
        plt.rcParams['ytick.labelsize'] = 17
        plt.rcParams['lines.linewidth'] = 3
        plt.rcParams['lines.markersize'] = 6
        plt.rcParams['font.size'] = 20
        plt.rcParams['axes.labelsize'] = 20

        plt.figure(0)
        plot = sns.histplot([self.D1, self.D2], stat='density',
                            alpha=0.6, kde=False, common_norm=False)
        plt.legend(labels=[f"{self.D2_name} (n={len(self.D2)})",
                           f"{self.D1_name} (n={len(self.D1)})"])
        plt.xlabel('Citations')

        plt.tight_layout()
        plt.savefig(f"results/{self.compare_name}_histogram.png")

    def normal_test(self):
        """Test on normal distribution via QQ-plot visualization."""

        plt.rcParams['figure.subplot.wspace'] = 0.6
        plt.rcParams['axes.titleweight'] = 'bold'
        plt.rcParams['axes.labelsize'] = 20

        plt.figure(figsize=(10, 4))
        ax1 = plt.subplot(121)
        stats.probplot(self.D1, dist="norm", plot=plt)
        ax1.text(0.5, 0.85, f"p = {round(stats.shapiro(self.D1).pvalue, 6)}",
                 transform=ax1.transAxes, ha='right', fontsize=16)
        plt.title(self.D1_name)

        ax2 = plt.subplot(122)
        stats.probplot(self.D2, dist="norm", plot=plt)
        ax2.text(0.5, 0.85, f"p = {round(stats.shapiro(self.D2).pvalue, 6)}",
                 transform=ax2.transAxes, ha='right', fontsize=16)
        plt.title(self.D2_name)

        plt.tight_layout()
        plt.savefig(f"results/{self.compare_name}_QQ_normal_distribution.png")
        plt.close()

    def t_distr_test(self):
        """Test on noncentral t distribution via QQ-plot visualization.

        Return:
            QQ-plot for each dataset assuming non-central t distribution
        """
        # Fit distribution parameters
        param_D1 = stats.nct.fit(self.D1)
        param_D2 = stats.nct.fit(self.D2)

        plt.rcParams['figure.subplot.wspace'] = 0.6
        plt.rcParams['axes.labelsize'] = 20

        plt.figure(figsize=(10, 4))
        ax1 = plt.subplot(121)
        stats.probplot(self.D1, dist="nct", sparams=(param_D1), plot=plt)
        stat_D1, p_val_D1 = stats.kstest(
            self.D1, 'nct', args=(param_D1), N=200,
            alternative='two-sided', mode='auto')
        ax1.text(0.96, 0.1, 'p = ' + str(round(p_val_D1, 3)),
                 transform=ax1.transAxes, ha='right', fontsize=20)
        plt.title(self.D1_name)

        ax2 = plt.subplot(122)
        stats.probplot(self.D2, dist="nct", sparams=(param_D2), plot=plt)
        stat_D2, p_val_D2 = stats.kstest(
            self.D2, 'nct', args=(param_D2), N=200, alternative='two-sided',
            mode='auto')
        ax2.text(0.96, 0.1, 'p = ' + str(round(p_val_D2, 3)),
                 transform=ax2.transAxes, ha='right', fontsize=20)
        plt.title(self.D2_name)

        plt.tight_layout()
        plt.savefig(f"results/{self.compare_name}_QQ_non-central_t_distribution.png")

    def bayesian_hypotheses_test(self):
        """Student-t distribution is used to discribe the data distributions
        (is less sensitive to outliers than the normal distribution)
        """
        self.pooled_data = np.concatenate((self.D1, self.D2), axis=None)

        pooled_mean = self.pooled_data.mean()
        pooled_sigma = self.pooled_data.std()

        with pymc.Model() as self.model:
            # ----------
            # Priors
            # Mean and standard deviation priors are drawn from a normal
            # or uniform distribution, respectively.
            # The priors are taken for each group separately,
            # but all from the pooled data (i.e. the data of both hypotheses),
            # to not imply the desired outcome from the beginning
            # but have a flat prior.
            # ----------
            D1_prior_mean = pymc.Normal(self.D1_name + " Mean",
                                        mu=pooled_mean, sigma=pooled_sigma*10)
            D2_prior_mean = pymc.Normal(self.D2_name + " Mean",
                                        mu=pooled_mean, sigma=pooled_sigma*10)

            # Uniform distribution describes the groups standard deviations
            #  /1000 and *1000 from Kruschke too broad
            D1_prior_std = pymc.Uniform(
                self.D1_name + " Std", lower=pooled_sigma/100,
                upper=pooled_sigma*100)
            D2_prior_std = pymc.Uniform(
                self.D2_name + " Std", lower=pooled_sigma/100,
                upper=pooled_sigma*100)

            # v is an exponentially distributed prior for both hypotheses
            # for nu -> infty, shape is normal and for nu -> 0 havy-tailed
            ν = pymc.Exponential("Shape", 1 / 29.0)

            λ0 = D1_prior_std ** -2
            λ1 = D2_prior_std ** -2

            group1 = pymc.StudentT(self.D1_name, nu=ν+0.01,  # +1 in Kruschke
                                   mu=D1_prior_mean, lam=λ0, observed=self.D1)
            group2 = pymc.StudentT(self.D2_name, nu=ν+0.01,
                                   mu=D2_prior_mean, lam=λ1, observed=self.D2)

            diff_of_means = pymc.Deterministic("Difference of Means",
                                               D1_prior_mean - D2_prior_mean)
            diff_of_stds = pymc.Deterministic("Difference of Stds",
                                              D1_prior_std - D2_prior_std)
            effect_size = pymc.Deterministic(
                "Effect Size",
                diff_of_means
                / np.sqrt(((self.D1_n - 1) * D1_prior_std ** 2
                           + (self.D2_n - 1) * D2_prior_std ** 2)
                          / (self.D1_n + self.D2_n - 2))
            )  # Cohens d effect size with different sample sizes

            self.trace = pymc.sample(100000, tune=3000,
                                     return_inferencedata=True)

        # save trace
        self.trace.to_netcdf(f"results/traces/{self.compare_name}_trace.nc")
        # load via az.from_netcdf("filename.nc")

    def plot_bayesian_hypotheses_test(self):
        """Visualization of bayesian hypothese testing.
        Return:
            posterior plot : Posterior distributions of the mean, sigma and
                             shape parameters of for both datasets
            evaluation plot : Difference of means and standard deviations
                              and the effect size
        """

        plt.rcParams['xtick.labelsize'] = 25  # Set global parameters
        plt.rcParams['ytick.labelsize'] = 25
        plt.rcParams['lines.linewidth'] = 5
        plt.rcParams['lines.markersize'] = 25
        plt.rcParams["font.weight"] = "normal"
        plt.rcParams["axes.labelweight"] = "normal"
        plt.rcParams['axes.titleweight'] = 'bold'
        plt.rcParams['axes.titlesize'] = 17

        posterior_dict = {0: r'$\mu_{1}$'.replace('1', self.D1_name),
                          1: r'$\mu_{2}$'.replace('2', self.D2_name),
                          2: r'$\sigma_{1}$'.replace('1', self.D1_name),
                          3: r'$\sigma_{2}$'.replace('2', self.D2_name),
                          4: r'$\nu$',
                          5: ' '}
        evaluation_dict = {0: r'$\mu_{1} - \mu_{2}$'.replace('1', self.D1_name).replace('2', self.D2_name),
                           1: r'$\sigma_{1} -\sigma_{2}$'.replace('1', self.D1_name).replace('2', self.D2_name),
                           2: (r"$(\mu_{g1} - \mu_{g2})/"
                               r"\sqrt{\frac{(n_{g1}-1)\cdot \sigma_{g1}^2 + (n_{g2}-1)\cdot \sigma_{g2}^2}{n_{g1}+n_{g2}-2}}$"
                               .replace('g1', self.D1_name).replace('g2', self.D2_name))}

        # Use labeller to remove plot titles
        labeller = az.labels.MapLabeller(var_name_map={
            self.D1_name + " Mean": "",
            self.D2_name + " Mean": "",
            self.D1_name + " Std": "",
            self.D2_name + " Std": "",
            "Shape": ""})  # Removes variable names in plot

        fig, ax_obj = plt.subplots(1, 5, figsize=(5*5, 4))
        ax1 = az.plot_posterior(
            self.trace,
            var_names=[self.D1_name + " Mean", self.D2_name + " Mean",
                       self.D1_name + " Std", self.D2_name + " Std", "Shape"],
            hdi_prob=0.95,
            textsize=32,
            color="#87ceeb",
            labeller=labeller,
            ax=ax_obj,
        )
        for i, ax in enumerate(ax1.flat):
            # ax.text(-0.01, 1.12, letter_dict[i], transform=ax.transAxes,
            # fontsize=25, fontweight='bold', va='top', ha='right')
            ax.set_xlabel(posterior_dict[i], fontsize=34)

        plt.tight_layout()
        plt.savefig("results/" + self.compare_name + "_posterior.png", bbox_inches='tight')
        plt.show()

        # Use labeller to remove plot titles
        labeller2 = az.labels.MapLabeller(var_name_map={
            "Difference of Means": "",
            "Difference of Stds": "",
            "Effect Size": ""})

        fig, ax_obj2 = plt.subplots(1, 3, figsize=(6*3, 5))
        ax2 = az.plot_posterior(
            self.trace,
            var_names=["Difference of Means", "Difference of Stds", "Effect Size"],
            textsize=30,
            # rope=(-0.01, 0.01),
            ref_val=0,
            hdi_prob=0.95,
            color="#87ceeb",
            labeller=labeller2,
            ax=ax_obj2,
        )
        for i, ax in enumerate(ax2.flat):
            # ax.text(-0.01, 1.12, letter_dict[i], transform=ax.transAxes,
            # fontsize=25, fontweight='bold', va='top', ha='right')
            ax.set_xlabel(evaluation_dict[i], fontsize=30)

        plt.tight_layout()
        plt.savefig(f"results/{self.compare_name}_evaluation.png")
        plt.show()

    def plot_scatter(self):
        """Use arviz to plot a 2D scatterplot and visulize correlations."""

        plt.rcParams['lines.markersize'] = 9
        plt.rcParams['lines.linewidth'] = 2

        ax = az.plot_pair(
            self.trace,
            group='posterior',
            var_names=[self.D1_name + " Mean", self.D2_name + " Mean",
                       self.D1_name + " Std", self.D2_name + " Std", "Shape"],
            kind=['scatter'],
            scatter_kwargs={'alpha': 0.5},
            textsize=45,
        )
        for i, ax in enumerate(ax.flat):  # rotate y labels
            ax.set_ylabel(ax.get_ylabel(), rotation=30, ha='right')

        plt.tight_layout()
        plt.savefig(f"results/{self.compare_name}_2D_scatter.png")
        plt.close()

        plt.rcParams['lines.markersize'] = 4
        ax2 = az.plot_pair(
            self.trace,
            group='posterior',
            var_names=["Difference of Means", "Difference of Stds", "Effect Size"],
            # divergences=True,
            kind=['scatter', 'kde'],
            textsize=23,
        )
        plt.tight_layout()
        plt.savefig(f"results/{self.compare_name}_2D_scatter_differences_effect_size.png")
        plt.close()

    def posterior_predictive_check(self):
        """Use pymc to create posterior predictive samples
        and arviz to visulize if the posterior parameter predictions
        are in accordance with the observed data.
        """
        # Set global parameters for the plot
        plt.rcParams['figure.subplot.wspace'] = 0.2
        plt.rcParams['figure.figsize'] = (10, 6)
        plt.rcParams['font.size'] = 40
        plt.rcParams['xtick.labelsize'] = 40
        plt.rcParams['ytick.labelsize'] = 40
        plt.rcParams['axes.labelsize'] = 40

        # thinn chain for posterior predictive calculation - keep 1 of 10
        thinned_trace = self.trace.sel(draw=slice(None, None, 10))

        pymc.sample_posterior_predictive(thinned_trace, model=self.model,
                                         extend_inferencedata=True)

        ax1 = az.plot_ppc(thinned_trace, num_pp_samples=1000,
                          kind='kde', mean=False)

        for i, ax in enumerate(ax1.flat):
            ax.set_xlim([-30, 600])

        plt.tight_layout()
        plt.savefig(f"results/{self.compare_name}_posterior_predictive_check.png")

        # Cumulative
        ax2 = az.plot_ppc(thinned_trace, num_pp_samples=1000,
                          kind='cumulative', mean=True)

        for i, ax in enumerate(ax2.flat):
            ax.set_xlim([-30, 600])

        plt.tight_layout()
        plt.savefig(f"results/{self.compare_name}_posterior_predictive_check_cumulative.png")
        plt.close()

    def prior_predictive_check(self):
        """Draw from prior and translates prior parameters to observables
        (i.e. model evaluations with the drawn priors).
        A prior predicitve check is basically a posterior predictive check
        with the prior predictive dist. instead of the posterior predictive,
        i.e. without taking observed data into account.
        """

        # Set global parameters for the plot
        plt.rcParams['figure.subplot.wspace'] = 0.2
        plt.rcParams['figure.figsize'] = (10, 6)
        plt.rcParams['font.size'] = 40
        plt.rcParams['xtick.labelsize'] = 40
        plt.rcParams['ytick.labelsize'] = 40
        plt.rcParams['axes.labelsize'] = 40

        with self.model:
            prior_predicitve_samples = pymc.sample_prior_predictive(5000)

        ax1 = az.plot_ppc(prior_predicitve_samples, num_pp_samples=1000,
                          kind='kde', mean=False, group='prior')
        for i, ax in enumerate(ax1.flat):
            ax.set_xlim([-300, 400])
            # Plot legend outside of plot
            if i == (len(ax1.flat)-1):
                ax.legend(bbox_to_anchor=(1.02, 1), loc=2, borderaxespad=0.)
            else:
                ax.get_legend().remove()
        plt.tight_layout()
        plt.savefig(f"results/{self.compare_name}_prior_predictive_check.png")

        ax2 = az.plot_ppc(prior_predicitve_samples, num_pp_samples=1000,
                          kind='cumulative', mean=True, group='prior')
        for i, ax in enumerate(ax2.flat):
            ax.set_xlim([-300, 400])
            # Plot legend outside of plot
            if i == (len(ax1.flat)-1):
                ax.legend(bbox_to_anchor=(1.02, 1), loc=2, borderaxespad=0.)
            else:
                ax.get_legend().remove()
        plt.tight_layout()
        plt.savefig(f"results/{self.compare_name}_prior_predictive_check_cumulative.png")
        plt.close()

    def diagnostics(self):
        """Diagnostics of MCMC chain to check convergence"""

        # ----------------------------------
        # Calculate Potential Scale Reduction Factors (PSRF or rhat)
        # ----------------------------------
        PSRF = az.rhat(self.trace)
        # print(f"The PSRFs of {self.compare_name} are {PSRF}")
        # Save PSRFs, load with xarray.load_dataset()
        PSRF.to_netcdf(f"results/diagnostics/{self.compare_name}_PSRF.nc")

        # ----------------------------------
        # Plot trace for visual examination
        # ----------------------------------
        ax0 = az.plot_trace(self.trace, divergences="bottom")

        plt.tight_layout()
        plt.savefig(f"results/diagnostics/{self.compare_name}_traces.png")
        plt.close()

        # ----------------------------------
        # Calculate and visulize Effective Sample Size (ESS)
        # ----------------------------------
        ESS = az.ess(self.trace)
        # print(f"ESS of {self.compare_name} was {ESS}")
        ESS.to_netcdf(f"results/diagnostics/{self.compare_name}_ESS.nc")

        ax1 = az.plot_ess(self.trace, kind="local")

        plt.tight_layout()
        plt.savefig(f"results/diagnostics/{self.compare_name}_ESS_local.png")
        plt.close()




class Bayesian_multi_comparison:
    """Class for bayesian hypothese testing with more than two groups.
    The group data is assinged to the class.
    BEST method of Kruschke 2012, extended for multiple groups.
    Each group gets a different mu and sigma but the same nu.
    """

    def __init__(self, data: dict, compare_name: str):
        """Data of the hypotheses that are tested.
        Input:
            data : Dictionary containing the datasets to compare
            compare_name : name of the compare made (appears in plot names)
        """

        self.D = data
        self.compare_name = compare_name

        # Test if data is in the right format
        if not isinstance(list(self.D.values())[0], (np.ndarray, np.generic)):
            raise TypeError('Hypothesis data must be of type numpy ndarray.')

    def histogram_compare(self):
        """Plots seaborn histogram of the different groups"""
        import seaborn as sns

        # Set global parameters
        plt.rcParams['figure.figsize'] = (12, 7)  # Set global parameters
        plt.rcParams['xtick.labelsize'] = 17
        plt.rcParams['ytick.labelsize'] = 17
        plt.rcParams['lines.linewidth'] = 3
        plt.rcParams['lines.markersize'] = 6
        plt.rcParams['font.size'] = 20
        plt.rcParams['axes.labelsize'] = 20

        fig, ax1 = plt.subplots()
        sns.histplot([dataset for key, dataset in self.D.items()],
                     alpha=0.5, kde=True)
        ax1.set_xlabel('value')
        # create legend
        legend_handles = ax1.get_legend().legendHandles
        # be careful with the legend - seaborn messes up the order in arrays

        plt.savefig(f"results_multi/{self.compare_name}_histogram.png")
        plt.show()

    def normal_test(self):
        """Test on normal distribution via QQ-plot visualization."""

        plt.rcParams['figure.subplot.wspace'] = 0.6

        fig, ax = plt.subplots(len(self.D), 1, figsize=(5, 5*len(self.D)))
        for axs, key in zip(ax.ravel(), self.D.keys()):
            current_data = self.D[key]
            stats.probplot(current_data, dist="norm", plot=axs)
            axs.text(0.98, 0.1,
                     'p = ' + str(round(stats.shapiro(current_data).pvalue, 4)),
                     transform=axs.transAxes, ha='right', fontsize=16)
            axs.set_title(key)

        plt.tight_layout()
        plt.savefig(f"results_multi/{self.compare_name}_QQ_normal_distribution.png")
        plt.show()

    def t_distr_test(self):
        """Test on noncentral t distribution via QQ-plot visualization.
        Input:
            dataset dict
        Return:
            QQ-plot for each dataset assuming non-central t distribution
        """
        plt.rcParams['figure.subplot.wspace'] = 0.6

        fig, ax = plt.subplots(int(len(self.D)/2), 2,
                               figsize=(11, 5/2*len(self.D)))
        for axs, key in zip(ax.ravel(), self.D.keys()):
            current_data = self.D[key]
            # Fit distribution parameters
            distr_params = stats.nct.fit(current_data)

            stats.probplot(current_data, dist="nct",
                           sparams=(distr_params), plot=axs)
            stat, p_val = stats.kstest(current_data, 'nct',
                                       args=(distr_params), N=200,
                                       alternative='two-sided', mode='auto')
            axs.text(0.96, 0.1, f"p = {str(round(p_val, 3))}",
                     transform=axs.transAxes, ha='right', fontsize=16)
            axs.set_title(key)

        plt.tight_layout()
        plt.savefig(f"results_multi/{self.compare_name}_QQ_non-central_t_distribution.png")
        plt.show()

    def bayesian_hypotheses_test(self):
        """Student-t distribution is used to discribe the data distributions
        (is less sensitive to outliers than the normal distribution)
        """
        self.pooled_data = np.concatenate([data for key, data in self.D.items()], axis=None)

        pooled_mean = self.pooled_data.mean()
        pooled_sigma = self.pooled_data.std()

        with pymc.Model() as self.model:
            # ----------
            # Priors
            # Mean and standard deviation priors are drawn from a
            # normal or uniform distribution, respectively.
            # The priors are taken for each group separately,
            # but all from the pooled data (i.e. the data of all groups)
            # ----------

            # dicts to store prior values
            prior_means = {}
            prior_stds = {}
            # dict for group values
            groups_dict = {}

            # v = exponentially distributed prior - common prior for all groups
            ν = pymc.Exponential("Shape Parameter", 1 / 29.0)

            for key in self.D.keys():
                prior_means[key + ' Mean'] = pymc.Normal(
                    key + ' Mean', mu=pooled_mean, sigma=pooled_sigma*10)

                # Uniform distribution describes the groups standard deviations
                # for not too boroad range /100*100 instead of /1000 and *1000
                prior_stds[key + ' Std'] = pymc.Uniform(key + ' Std',
                                                        lower=pooled_sigma/100,
                                                        upper=pooled_sigma*100)

                groups_dict[key] = pymc.StudentT(
                    key, nu=ν + 1, mu=prior_means[key + ' Mean'],
                    lam=prior_stds[key + ' Std'] ** -2, observed=self.D[key])

            for key in self.D.keys():
                for key2 in self.D.keys():
                    current_diff_of_mean = pymc.Deterministic(
                        "Difference of Means" + key + '-'
                        + key2, prior_means[key + ' Mean']
                        - prior_means[key2 + ' Mean'])
                    current_diff_of_stds = pymc.Deterministic(
                        "Difference of Stds" + key + '-'
                        + key2, prior_stds[key + ' Std']
                        - prior_stds[key2 + ' Std'])

                    current_effect_size = pymc.Deterministic(
                        "Effect Size" + key + '-' + key2,
                        current_diff_of_mean
                        / np.sqrt(((len(self.D[key]) - 1)*prior_stds[key + ' Std']**2
                                  + (len(self.D[key2]) - 1)*prior_stds[key2 + ' Std']**2)
                                 /(len(self.D[key]) + len(self.D[key2]) - 2)))

            self.trace = pymc.sample(100000,
                                     tune=3000,
                                     return_inferencedata=True)

        # save trace
        self.trace.to_netcdf(f"results_multi/{self.compare_name}_trace.nc")

    def heatmap(self):
        """Visualization of bayesian hypothese testing.
        Return:
            A heatmap for the means of the difference of means,
            effect sizes and percentage of values above 0 for each comparison.
        """
        # Set global parameters
        plt.rcParams['xtick.labelsize'] = 20
        plt.rcParams['ytick.labelsize'] = 20

        keys = [key for key, values in self.D.items()]

        mean_diff_of_means = [[] for key in keys]
        mean_effect_size = [[] for key in keys]
        percentage_higher_0 = [[] for key in keys]

        # store mean values of compares in numpy array of arrays,
        # each array is a compare of one key with every other keys
        for key1, key_number in zip(keys, range(len(keys))):
            for key2 in keys:
                mean_diff_of_means[key_number].append(
                    self.trace.posterior["Difference of Means"
                                         + key1 + '-' + key2].mean())
                mean_effect_size[key_number].append(
                    self.trace.posterior["Effect Size" + key1 + '-' + key2].mean())
                percentage_higher_0[key_number].append(
                    (self.trace.posterior["Difference of Means" + key1 + '-' + key2] > 0).mean())

        titles = ["Mean Difference of Means", "Mean Effect Size",
                  "Difference of Means larger 0"]
        ndarrays = [mean_diff_of_means, mean_effect_size, percentage_higher_0]

        # ---------------------------------------------------------------------
        # Plot heatmaps
        # ---------------------------------------------------------------------
        for title, ndarray in zip(titles, ndarrays):

            fig, ax = plt.subplots(figsize=(len(keys), len(keys)))
            colormap = sns.color_palette("Reds")
            if title == "Difference of Means larger 0":
                im = sns.heatmap(ndarray, cmap=colormap, annot=True,
                        cbar_kws={"shrink": 0.8},
                        square=True, fmt=".1%", annot_kws={"fontsize":13})
            else:
                im = sns.heatmap(ndarray, cmap=colormap, annot=True,
                        cbar_kws={"shrink": 0.8},
                        square=True, annot_kws={"fontsize":13})

            # Show all ticks and label them with the respective list entries
            ax.set_xticks(np.arange(len(keys)), labels=keys)
            ax.set_yticks(np.arange(len(keys)), labels=keys)

            # Rotate the tick labels and set their alignment.
            plt.setp(ax.get_xticklabels(), rotation=45, ha="right", va="top",
                     rotation_mode="anchor")
            plt.setp(ax.get_yticklabels(), rotation=45, ha="right", va="top",
                     rotation_mode="anchor")

            ax.set_title(title)
            fig.tight_layout()
            plt.savefig(f"results_multi/{self.compare_name}_{title}heatmap.png")
            
            # only show heatmap with percentage_higher_0
            if title == "Difference of Means larger 0":
                plt.show()
            else:
                plt.close()
            
    def heatmap_citation(self):
        """Visualization of bayesian hypothese testing.
        Specialized visualization function for citation data with NR and R 
        ending.
        Here less compares are shown.
        Return:
            A heatmap for the means of the difference of means,
            effect sizes and percentage of values above 0 for each comparison.
        """
        # Set global parameters
        plt.rcParams['xtick.labelsize'] = 30
        plt.rcParams['ytick.labelsize'] = 30

        keys = [key for key, values in self.D.items()]
        keyR = [key for key in keys if key.endswith(' R')]
        keyNR = [key for key in keys if key.endswith(' NR')]

        mean_diff_of_means = [[] for key in keyR]
        mean_effect_size = [[] for key in keyR]
        percentage_higher_0 = [[] for key in keyR]

        # store mean values of compares in numpy array of arrays,
        # each array is a compare of one key with every other keys
        for key1, key_number in zip(keyR, range(len(keyR))):
            for key2 in keyNR:
                mean_diff_of_means[key_number].append(
                    self.trace.posterior["Difference of Means"
                                         + key1 + '-' + key2].mean())
                mean_effect_size[key_number].append(
                    self.trace.posterior["Effect Size" + key1 + '-' + key2].mean())
                percentage_higher_0[key_number].append(
                    (self.trace.posterior["Difference of Means" + key1 + '-' + key2] > 0).mean())

        titles = ["Mean Difference of Means", "Mean Effect Size",
                  "Difference of Means larger 0"]
        ndarrays = [mean_diff_of_means, mean_effect_size, percentage_higher_0]

        # ---------------------------------------------------------------------
        # Plot heatmaps
        # ---------------------------------------------------------------------
        for title, ndarray in zip(titles, ndarrays):

            fig, ax = plt.subplots(figsize=(len(keys), len(keys)))
            colormap = sns.color_palette("Reds")
            if title == "Difference of Means larger 0":
                im = sns.heatmap(ndarray, cmap=colormap, annot=True,
                        cbar_kws={"shrink": 0.8},
                        square=True, fmt=".1%", annot_kws={"fontsize":20})
            else:
                im = sns.heatmap(ndarray, cmap=colormap, annot=True,
                        cbar_kws={"shrink": 0.8},
                        square=True, annot_kws={"fontsize":20})

            # Show all ticks and label them with the respective list entries
            ax.set_xticks(np.arange(len(keyR)), labels=keyR)
            ax.set_yticks(np.arange(len(keyNR)), labels=keyNR)

            # Rotate the tick labels and set their alignment.
            plt.setp(ax.get_xticklabels(), rotation=45, ha="right", va="top",
                     rotation_mode="anchor")
            plt.setp(ax.get_yticklabels(), rotation=45, ha="right", va="top",
                     rotation_mode="anchor")

            #ax.set_title(title)
            fig.tight_layout()
            plt.savefig(f"results_multi/{self.compare_name}_{title}heatmap_small.png")
            plt.show()

    def plot_posteriors(self):
        import math
        # Set global parameters for the plot
        plt.rcParams['xtick.labelsize'] = 25
        plt.rcParams['ytick.labelsize'] = 25
        plt.rcParams['lines.linewidth'] = 7
        plt.rcParams['lines.markersize'] = 25
        plt.rcParams["font.weight"] = "normal"
        plt.rcParams["axes.labelweight"] = "normal"
        plt.rcParams['axes.titleweight'] = 'bold'

        # Matrix plot for differences of means
        fig, ax = plt.subplots(
            int(math.sqrt(len(self.D)**2)), int(math.sqrt(len(self.D)**2)),
            figsize=(4*int(math.sqrt(len(self.D)**2)),
                     4*int(math.sqrt(len(self.D)**2))))

        for axs, key in zip(ax.ravel(), [key1 + '-' + key2 for key1 in self.D.keys() for key2 in self.D.keys()]):

            if key in [key1 + '-' + key1 for key1 in self.D.keys()]:
                continue  # skip self compares in plotting
            axs = az.plot_posterior(
                self.trace,
                var_names=["Difference of Means" + key],
                ref_val=0,
                hdi_prob=0.95,
                color="#87ceeb",
                ax=axs
            )
            axs.set_title(key, fontsize=27)
            axs.set_xlabel(r'$\mu_1 - \mu_2$', fontsize=35)

        plt.tight_layout()
        plt.savefig(f"results_multi/{self.compare_name}"
                    "_difference_of_mean_posterior_matrix.png")

        # Matrix plot for effect size
        fig, ax = plt.subplots(
            int(math.sqrt(len(self.D)**2)), int(math.sqrt(len(self.D)**2)),
            figsize=(4.2*int(math.sqrt(len(self.D)**2)),
                     4.2*int(math.sqrt(len(self.D)**2))))

        for axs, key in zip(ax.ravel(), [key1 + '-' + key2 for key1 in self.D.keys() for key2 in self.D.keys()]):
            if key in [key1 + '-' + key1 for key1 in self.D.keys()]:
                continue  # skip self compares in plotting
            axs = az.plot_posterior(
                self.trace,
                var_names=["Effect Size" + key],
                ref_val=0,
                hdi_prob=0.95,
                color="#87ceeb",
                ax=axs,
            )
            axs.set_title(key, fontsize=27)
            axs.set_xlabel(
                r"$(\mu_1 - \mu_2)/ \sqrt{\frac{(n_1-1)\cdot"
                r"\sigma_1^2 + (n_2-1)\cdot \sigma_2^2}{n_1+n_2-2}}$",
                fontsize=20)

        plt.tight_layout()
        plt.savefig(f"results_multi/{self.compare_name}_effect_size_posterior_matrix.png") 

    def prior_predictive_check(self, xlim=None):
        """Draw from prior and translates prior parameters to observables
        (i.e. model evaluations with the drawn priors).
        A prior predicitve check is basically a posterior predictive check
        with the prior predictive distr. instead of the posterior predictive,
        i.e. without taking observed data into account.
        Input:
            xlim : x-limits for posterior plot (optional)
        """

        # Set global parameters for the plot
        plt.rcParams['figure.subplot.wspace'] = 0.2
        plt.rcParams['figure.figsize'] = (10, 6)
        plt.rcParams['font.size'] = 40
        plt.rcParams['xtick.labelsize'] = 40
        plt.rcParams['ytick.labelsize'] = 40
        plt.rcParams['axes.labelsize'] = 40

        with self.model:
            prior_predicitve_samples = pymc.sample_prior_predictive(5000)

        ax1 = az.plot_ppc(prior_predicitve_samples, num_pp_samples=1000,
                          kind='kde', mean=False, group='prior')
        for i, ax in enumerate(ax1.flat):
            if xlim is not None:
                ax.set_xlim(xlim)
            # Plot legend outside of plot
            if i == (len(ax1.flat)-1):
                ax.legend(bbox_to_anchor=(1.02, 1), loc=2, borderaxespad=0.)
            else:
                ax.get_legend().remove()
        plt.tight_layout()
        plt.savefig(f"results_multi/{self.compare_name}_prior_predictive_check.png")

        ax2 = az.plot_ppc(prior_predicitve_samples, num_pp_samples=1000,
                          kind='cumulative', mean=True, group='prior')
        for i, ax in enumerate(ax2.flat):
            if xlim is not None:
                ax.set_xlim(xlim)
            # Plot legend outside of plot
            if i == (len(ax1.flat)-1):
                ax.legend(bbox_to_anchor=(1.02, 1), loc=2, borderaxespad=0.)
            else:
                ax.get_legend().remove()
        plt.tight_layout()
        plt.savefig(f"results_multi/{self.compare_name}_prior_predictive_check_cumulative.png")
        plt.close()

    def posterior_predictive_check(self, xlim=None):
        """Use pymc to create posterior predictive samples
        and arviz to visulize if the posterior parameter predictions
        are in accordance with the observed data.
        Input:
            xlim : x-limits for posterior plot (optional)
        """
        # Set global parameters for the plot
        plt.rcParams['figure.subplot.wspace'] = 0.2
        plt.rcParams['figure.figsize'] = (10, 6)
        plt.rcParams['font.size'] = 40
        plt.rcParams['xtick.labelsize'] = 40
        plt.rcParams['ytick.labelsize'] = 40
        plt.rcParams['axes.labelsize'] = 40

        # thinn chain for posterior predictive calculation - keep 1 of 10
        thinned_trace = self.trace.sel(draw=slice(None, None, 10))

        pymc.sample_posterior_predictive(thinned_trace, model=self.model,
                                         extend_inferencedata=True)

        ax1 = az.plot_ppc(thinned_trace, num_pp_samples=1000,
                          kind='kde', mean=False)

        if xlim is not None:
            for i, ax in enumerate(ax1.flat):
                ax.set_xlim(xlim)

        plt.tight_layout()
        plt.savefig(f"results_multi/{self.compare_name}_posterior_predictive_check.png")

        # Cumulative
        ax2 = az.plot_ppc(thinned_trace, num_pp_samples=1000,
                          kind='cumulative', mean=True)

        if xlim is not None:
            for i, ax in enumerate(ax2.flat):
                ax.set_xlim(xlim)

        plt.tight_layout()
        plt.savefig(f"results_multi/{self.compare_name}_posterior_predictive_check_cumulative.png")
        plt.close()


    def diagnostics(self):
        """Diagnostics of MCMC chain to check convergence"""

        # ----------------------------------
        # Calculate Potential Scale Reduction Factors (PSRF or rhat)
        # ----------------------------------
        PSRF = az.rhat(self.trace)
        # print(f"The PSRFs of {self.compare_name} are {PSRF}")
        # Save PSRFs, load with xarray.load_dataset()
        PSRF.to_netcdf(f"results_multi/diagnostics/{self.compare_name}_PSRF.nc")

        # ----------------------------------
        # Plot trace for visual examination
        # ----------------------------------
        ax0 = az.plot_trace(self.trace, divergences="bottom")

        plt.tight_layout()
        plt.savefig(f"results_multi/diagnostics/{self.compare_name}_traces.png")
        plt.close()

        # ----------------------------------
        # Calculate and visulize Effective Sample Size (ESS)
        # ----------------------------------
        ESS = az.ess(self.trace)
        # print(f"ESS of {self.compare_name} was {ESS}")
        ESS.to_netcdf(f"results_multi/diagnostics/{self.compare_name}_ESS.nc")

        ax1 = az.plot_ess(self.trace, kind="local")

        plt.tight_layout()
        plt.savefig(f"results_multi/diagnostics/{self.compare_name}_ESS_local.png")
        plt.close()
